# -*- coding: utf-8 -*-

import numpy as np

import leibniz.core.basis.theta_phi_r as delegate


_name_ = 'lng,lat,alt'
_params_ = ('lng', 'lat', 'alt')


def d2r(d):
    return d * np.pi / 180


def r2d(r):
    return r / np.pi * 180


def transform(lng, lat, alt, r0=6371000, device=-1):
    theta, phi, r = d2r(lng), d2r(lat), alt + r0
    return delegate.transform(theta, phi, r)


def dtransform(lng, lat, alt, dlng, dlat, dalt, r0=6371000, device=-1):
    theta, phi, r = d2r(lng), d2r(lat), alt + r0
    dtheta, dphi, dr = d2r(dlng), d2r(dlat), dalt

    return delegate.dtransform(theta, phi, r, dtheta, dphi, dr)


def inverse(x, y, z, r0=6371000, device=-1):
    theta, phi, r = delegate.inverse(x, y, z)
    lng, lat, alt = r2d(theta), r2d(phi), r - r0

    return lng, lat, alt


def dinverse(x, y, z, dx, dy, dz, **kwargs):
    dtheta, dphi, dr = delegate.dinverse(x, y, z, dx, dy, dz)
    dlng, dlat, dalt = r2d(dtheta), r2d(dphi), dr

    return dlng, dlat, dalt



# -*- coding: utf-8 -*-

import numpy as np
import torch as th
import leibniz as lbnz

from cached_property import cached_property
from torch import Tensor
from leibniz import cast


class RegularGrid:
    def __init__(self, basis, L=2, W=2, H=2, east=1, west=-1, north=1, south=-1, upper=1, lower=-1):
        self.basis = basis

        self.L = L
        self.W = W
        self.H = H
        self.shape = (1, 1, self.L, self.W, self.H)

        self.east = east
        self.west = west
        self.north = north
        self.south = south
        self.upper = upper
        self.lower = lower

        self.default_device = -1

    def get_device(self):
        return self.default_device

    def set_device(self, ix):
        self.default_device = ix

    def mk_zero(self) -> Tensor:
        return cast(np.zeros(self.shape), device=self.default_device)

    @cached_property
    def zero(self) -> Tensor:
        return self.mk_zero()

    def mk_one(self) -> Tensor:
        return cast(np.ones(self.shape), device=self.default_device)

    @cached_property
    def one(self) -> Tensor:
        return self.mk_one()

    def boundary(self) -> Tensor:
        foreward = self.mk_zero()
        backward = self.mk_zero()
        backward[-1, :, :] = 1
        foreward[0, :, :] = 1

        top = self.mk_zero()
        bottom = self.mk_zero()
        top[:, :, -1] = 1
        bottom[:, :, 0] = 1

        left = self.mk_zero()
        right = self.mk_zero()
        right[:, -1, :] = 1
        left[:, 0, :] = 1

        return th.cat([
            top, bottom, foreward, backward, left, right
        ], dim=1).view(1, 6, self.W, self.L, self.H)

    def data(self) -> Tensor:
        return cast(np.meshgrid(
            np.linspace(self.west, self.east, num=self.L),
            np.linspace(self.south, self.north, num=self.W),
            np.linspace(self.lower, self.upper, num=self.H), indexing='ij'
        ), device=self.default_device).view(1, 3, self.L, self.W, self.H)

    def delta(self) -> Tensor:
        return th.cat([
            (self.east - self.west) / (self.L - 1) * self.one,
            (self.north - self.south) / (self.W - 1) * self.one,
            (self.upper - self.lower) / (self.H - 1) * self.one
        ], dim=1).view(1, 3, self.L, self.W, self.H)

    def random(self) -> Tensor:
        d = lbnz.get_device()
        r = th.rand(1, 1, self.L, self.W, self.H, dtype=th.float64)
        if th.cuda.is_available():
            return r.cuda(device=d)
        else:
            return r



# -*- coding: utf-8 -*-

import torch as th


def add(x, y):
    x1, x2, x3 = x
    y1, y2, y3 = y

    return x1 + y1, x2 + y2, x3 + y3


def sub(x, y):
    x1, x2, x3 = x
    y1, y2, y3 = y

    return x1 - y1, x2 - y2, x3 - y3


def multiply(x, y):
    x1, x2, x3 = x
    y1, y2, y3 = y

    return x1 * y1, x2 * y2, x3 * y3


def dot(x, y):
    x1, x2, x3 = x
    y1, y2, y3 = y

    return x1 * y1 + x2 * y2 + x3 * y3


def cross(x, y):
    x1, x2, x3 = x
    y1, y2, y3 = y

    return (
        x2 * y3 - x3 * y2,
        x3 * y1 - x1 * y3,
        x1 * y2 - x2 * y1,
    )


def norm(v):
    return th.sqrt(th.sum(dot(v, v), dim=1, keepdim=True))


def normsq(v):
    return th.sum(dot(v, v), dim=1, keepdim=True)


def normalize(v):
    x, y, z = v
    r = th.sqrt(x * x + y * y + z * z)

    return x / r, y / r, z / r


def box(a, b, c):
    return dot(a, cross(b, c))


def det(transform):
    return box(transform[0:1, 0:3], transform[1:2, 0:3], transform[2:3, 0:3])
